@@ -2,7 +2,7 @@
 
 # pykar - KAR/MID Karaoke Player
 #
-# Copyright (C) 2007  Kelvin Lawson (kelvinl@users.sf.net)
+# Copyright (C) 2010 Kelvin Lawson (kelvinl@users.sf.net)
 #
 # This library is free software; you can redistribute it and/or
 # modify it under the terms of the GNU Lesser General Public
@@ -50,6 +50,12 @@
 # To play the MIDI songs on Linux, Timidity++ is also required:
 # . Timidity++ (timidity.sourceforge.net)
 
+# OSX REQUIREMENTS
+#
+# On OSX, pygame will run MIDI natively by default, but if the GUS
+# patches are installed in /usr/local/lib/timidity, it will run MIDI
+# via Timidity instead, which appears to work better than the native
+# support, so we recommend this.
 
 # USAGE INSTRUCTIONS
 #
@@ -72,10 +78,10 @@
 # and doneCallback:
 #
 # errorNotifyCallback, if provided, will be used to print out any error
-# messages (e.g. song file not found). This allows the module to fit 
+# messages (e.g. song file not found). This allows the module to fit
 # together well with GUI playlist managers by utilising the same GUI's
 # error popup window mechanism (or similar). If no callback is provided,
-# errors are printed to stdout. errorNotifyCallback should take one 
+# errors are printed to stdout. errorNotifyCallback should take one
 # parameter, the error string, e.g.:
 #   def errorPopup (ErrorString):
 #       msgBox (ErrorString)
@@ -144,7 +150,7 @@ import pygame, sys, os, struct, cStringIO
 VIEW_PERCENT = 33
 
 # Default font size at 480 pixels.
-FONT_SIZE = 50
+FONT_SIZE = 40
 
 # How much lead time before a new paragraph is scrolled up into view
 # (scrolling the old paragraph off), in milliseconds.  This only comes
@@ -163,16 +169,14 @@ debug = False
 class midiFile:
     def __init__(self):
         self.trackList = []         # List of TrackDesc track descriptors
-        self.text_events = Lyrics()       # Lyrics (0x1 events)
-        self.lyric_events = Lyrics()      # Lyrics (0x5 events)
 
         # Chosen lyric list from above.  It is converted by
         # computeTiming() from a list of (clicks, text) into a list of
         # (ms, text).
         self.lyrics = []
 
-        # self.text_charset = "iso-8859-13"
-        self.text_charset = ""      # The encoding of text in midi file
+        # self.text_encoding = "iso-8859-13"
+        self.text_encoding = ""      # The encoding of text in midi file
 
         self.ClickUnitsPerSMPTE = None
         self.SMPTEFramesPerSec = None
@@ -184,13 +188,14 @@ class midiFile:
         # all the tracks (and thus collected all the tempo changes),
         # we can go back and apply this knowledge to the other tracks.
         self.Tempo = [(0, 0)]
-        
+
         self.Numerator = None               # Numerator
         self.Denominator = None             # Denominator
         self.ClocksPerMetronomeTick = None  # MIDI clocks per metronome tick
         self.NotesPer24MIDIClocks = None    # 1/32 Notes per 24 MIDI clocks
-        self.earliestNoteMS = 0             # Click of earliest note in song
-        
+        self.earliestNoteMS = 0             # Start of earliest note in song
+        self.lastNoteMS = 0                 # End of latest note in song
+
 
 class TrackDesc:
     def __init__(self, trackNum):
@@ -199,15 +204,20 @@ class TrackDesc:
         self.BytesRead = 0              # Number of file bytes read for track
         self.FirstNoteClick = None      # Start of first note in track
         self.FirstNoteMs = None         # The same, in milliseconds
+        self.LastNoteClick = None       # End of last note in track
+        self.LastNoteMs = None          # In millseconds
         self.LyricsTrack = False        # This track contains lyrics
         self.RunningStatus = 0          # MIDI Running Status byte
 
+        self.text_events = Lyrics()       # Lyrics (0x1 events)
+        self.lyric_events = Lyrics()      # Lyrics (0x5 events)
+
 
 class MidiTimestamp:
     """ This class is used to apply the tempo changes to the click
     count, thus computing a time in milliseconds for any number of
     clicks from the beginning of the song. """
-    
+
     def __init__(self, midifile):
         self.ClickUnitsPerQuarter = midifile.ClickUnitsPerQuarter
         self.Tempo = midifile.Tempo
@@ -221,10 +231,10 @@ class MidiTimestamp:
         if clicks < 0:
             # Ignore jumps backward in time.
             return
-        
+
         while clicks > 0 and self.i < len(self.Tempo):
             # How many clicks remain at the current tempo?
-            clicksRemaining = self.Tempo[self.i][0] - self.click
+            clicksRemaining = max(self.Tempo[self.i][0] - self.click, 0)
             clicksUsed = min(clicks, clicksRemaining)
             if clicksUsed != 0:
                 self.ms += self.getTimeForClicks(clicksUsed, self.Tempo[self.i - 1][1])
@@ -243,7 +253,7 @@ class MidiTimestamp:
     def getTimeForClicks(self, clicks, tempo):
         microseconds = ( ( float(clicks) / self.ClickUnitsPerQuarter ) * tempo );
         time_ms = microseconds / 1000
-        return (time_ms)        
+        return (time_ms)
 
 class LyricSyllable:
     """ Each instance of this class records a single lyric event,
@@ -340,13 +350,32 @@ class Lyrics:
             # Paragraph break.  We treat it the same as line break,
             # but with an extra blank line.
             self.line += 2
-        
-        elif text == '\r':
+
+        elif text == '\r' or text == '\r\n':
             # Line break.
             self.line += 1
 
         elif text:
-            self.list.append(LyricSyllable(click, text, self.line))
+            text = text.replace('\r', '')
+
+            if text[0] == '\\':
+                # Paragraph break.  This is a text event convention, not a
+                # lyric event convention, but some midi files don't play
+                # by the rules.
+                self.line += 2
+                text = text[1:]
+            elif text[0] == '/':
+                # Line break.  A text convention, but see above.
+                self.line += 1
+                text = text[1:]
+
+            # Lyrics aren't supposed to include embedded newlines, but
+            # sometimes they do anyway.
+            lines = text.split('\n')
+            self.list.append(LyricSyllable(click, lines[0], self.line))
+            for line in lines[1:]:
+                self.line += 1
+                self.list.append(LyricSyllable(click, line, self.line))
 
     def computeTiming(self, midifile):
         # Walk through the lyrics and convert the click information to
@@ -363,6 +392,60 @@ class Lyrics:
             if track_desc.FirstNoteClick != None:
                 ts.advanceToClick(track_desc.FirstNoteClick)
                 track_desc.FirstNoteMs = ts.ms
+                if debug:
+                    print "T%s first note at %s clicks, %s ms" % (
+                        track_desc.TrackNum, track_desc.FirstNoteClick,
+                        track_desc.FirstNoteMs)
+            if track_desc.LastNoteClick != None:
+                ts.advanceToClick(track_desc.LastNoteClick)
+                track_desc.LastNoteMs = ts.ms
+
+    def analyzeSpaces(self):
+        """ Checks for a degenerate case: no (or very few) spaces
+        between words.  Sometimes Karaoke writers omit the spaces
+        between words, which makes the text very hard to read.  If we
+        detect this case, repair it by adding spaces back in. """
+
+        # First, group the syllables into lines.
+        lineNumber = None
+        lines = []
+        currentLine = []
+
+        for syllable in self.list:
+            if syllable.line != lineNumber:
+                if currentLine:
+                    lines.append(currentLine)
+                currentLine = []
+                lineNumber = syllable.line
+            currentLine.append(syllable)
+
+        if currentLine:
+            lines.append(currentLine)
+
+        # Now, count the spaces between the syllables of the lines.
+        totalNumSyls = 0
+        totalNumGaps = 0
+        for line in lines:
+            numSyls = len(line) - 1
+            numGaps = 0
+            for i in range(numSyls):
+                if line[i].text.rstrip() != line[i].text or \
+                   line[i + 1].text.lstrip() != line[i + 1].text:
+                    numGaps += 1
+
+            totalNumSyls += numSyls
+            totalNumGaps += numGaps
+
+        if totalNumSyls and float(totalNumGaps) / float(totalNumSyls) < 0.1:
+            # Too few spaces.  Insert more.
+            for line in lines:
+                for syllable in line[:-1]:
+                    if syllable.text.endswith('-'):
+                        # Assume a trailing hyphen means to join syllables.
+                        syllable.text = syllable.text[:-1]
+                    else:
+                        syllable.text += ' '
+
 
     def wordWrapLyrics(self, font):
         # Walks through the lyrics and folds each line to the
@@ -375,7 +458,7 @@ class Lyrics:
             return []
 
         maxWidth = manager.displaySize[0] - X_BORDER * 2
-        
+
         lines = []
 
         x = 0
@@ -388,7 +471,7 @@ class Lyrics:
             # displayed.
             syllable.left = None
             syllable.right = None
-            
+
             while lineNumber < syllable.line:
                 # A newline.
                 lines.append(currentLine)
@@ -453,16 +536,16 @@ class Lyrics:
     def write(self):
         # Outputs the lyrics, one line at a time.
         for syllable in self.list:
-            print "%s %s %s" % (syllable.ms, syllable.line, syllable.text)
-        
-def midiParseData(midiData, ErrorNotifyCallback, Charset):
-    
+            print "%s(%s) %s %s" % (syllable.ms, syllable.click, syllable.line, repr(syllable.text))
+
+def midiParseData(midiData, ErrorNotifyCallback, Encoding):
+
     # Create the midiFile structure
     midifile = midiFile()
-    midifile.text_charset = Charset
-    
+    midifile.text_encoding = Encoding
+
     # Open the file
-    filehdl = cStringIO.StringIO(midiData) 
+    filehdl = cStringIO.StringIO(midiData)
 
     # Check it's a MThd chunk
     packet = filehdl.read(8)
@@ -492,7 +575,9 @@ def midiParseData(midiData, ErrorNotifyCallback, Charset):
         # Check it's a MTrk
         ChunkType, Length = struct.unpack('>4sL', packet)
         if (ChunkType != "MTrk"):
-            print ("Didn't find expected MIDI Track")
+            if debug:
+                print ("Didn't find expected MIDI Track")
+
         # Process the track, getting a TrackDesc structure
         track_desc = midiParseTrack(filehdl, midifile, trackNum, Length, ErrorNotifyCallback)
         if track_desc:
@@ -507,37 +592,65 @@ def midiParseData(midiData, ErrorNotifyCallback, Charset):
     # Close the open file
     filehdl.close()
 
-    # Decide which list of lyric events to choose. There may be text events (0x01),
-    # lyric events (0x05) or sometimes both for compatibility. If both are
-    # available, we choose text events.
-    if midifile.text_events.hasAny() and midifile.lyric_events.hasAny():
-        midifile.lyrics = midifile.text_events
-        #midifile.lyrics = midifile.lyric_events
-    elif midifile.text_events.hasAny():
-        midifile.lyrics = midifile.text_events
-    elif midifile.lyric_events.hasAny():
-        midifile.lyrics = midifile.lyric_events
-    else:
+    # Get the lyrics from the best track.  We prefer any tracks that
+    # are "lyrics" tracks.  Failing that, we get the track with the
+    # most number of syllables.
+    bestSortKey = None
+    midifile.lyrics = None
+
+    for track_desc in midifile.trackList:
+        lyrics = None
+
+        # Decide which list of lyric events to choose. There may be
+        # text events (0x01), lyric events (0x05) or sometimes both
+        # for compatibility. If both are available, we choose the one
+        # with the most syllables, or text if they're the same.
+        if track_desc.text_events.hasAny() and track_desc.lyric_events.hasAny():
+            if len(track_desc.lyric_events.list) > len(track_desc.text_events.list):
+                lyrics = track_desc.lyric_events
+            else:
+                lyrics = track_desc.text_events
+        elif track_desc.text_events.hasAny():
+            lyrics = track_desc.text_events
+        elif track_desc.lyric_events.hasAny():
+            lyrics = track_desc.lyric_events
+
+        if not lyrics:
+            continue
+        sortKey = (track_desc.LyricsTrack, len(lyrics.list))
+        if sortKey > bestSortKey:
+            bestSortKey = sortKey
+            midifile.lyrics = lyrics
+
+    if not midifile.lyrics:
         ErrorNotifyCallback ("No lyrics in the track")
         return None
 
     midifile.lyrics.computeTiming(midifile)
+    midifile.lyrics.analyzeSpaces()
 
-    # Calculate the song start click (earliest note event in all tracks)
+    # Calculate the song start (earliest note event in all tracks), as
+    # well as the song end (last note event in all tracks).
     earliestNoteMS = None
+    lastNoteMS = None
     for track in midifile.trackList:
         if track.FirstNoteMs != None:
             if (track.FirstNoteMs < earliestNoteMS) or (earliestNoteMS == None):
                 earliestNoteMS = track.FirstNoteMs
+        if track.LastNoteMs != None:
+            if (track.LastNoteMs > lastNoteMS) or (lastNoteMS == None):
+                lastNoteMS = track.LastNoteMs
     midifile.earliestNoteMS = earliestNoteMS
+    midifile.lastNoteMS = lastNoteMS
 
     if debug:
         print "first = %s" % (midifile.earliestNoteMS)
+        print "last = %s" % (midifile.lastNoteMS)
 
     # Return the populated midiFile structure
     return midifile
-    
-    
+
+
 def midiParseTrack (filehdl, midifile, trackNum, Length, ErrorNotifyCallback):
     # Create the new TrackDesc structure
     track = TrackDesc(trackNum)
@@ -564,7 +677,7 @@ def midiProcessEvent (filehdl, track_desc, midifile, ErrorNotifyCallback):
     byteStr = filehdl.read(1)
     bytesRead = bytesRead + 1
     status_byte = ord(byteStr)
-    
+
     # Handle the MIDI running status. This allows consecutive
     # commands of the same event type to not bother sending
     # the event type again. If the top bit isn't set it's a
@@ -584,7 +697,7 @@ def midiProcessEvent (filehdl, track_desc, midifile, ErrorNotifyCallback):
         event_type = track_desc.RunningStatus
         filehdl.seek (-1, 1)
         bytesRead = bytesRead - 1
-    
+
     #print ("T%d: VarBytes = %d, event_type = 0x%X" % (track_desc.TrackNum, varBytes, event_type))
 ##     if debug:
 ##         print "Event: 0x%X" % event_type
@@ -608,24 +721,27 @@ def midiProcessEvent (filehdl, track_desc, midifile, ErrorNotifyCallback):
                 # Nothing left to discard
                 pass
             else:
-                print ("Invalid sequence number (%d)" % type)
+                if debug:
+                    print ("Invalid sequence number (%d)" % type)
         elif event == 0x01:
             # Text Event
             Length, varBytes = varLength(filehdl)
             bytesRead = bytesRead + varBytes
             text = filehdl.read(Length)
-            if (midifile.text_charset != "") :
-                try:
-                    text = text.decode(midifile.text_charset)
-                except :
-                    print "Unicode conversion error"
             bytesRead = bytesRead + Length
-            # Take out any Sysex text events, and append to the lyrics list
-            if (" SYX" not in text) and ("Track-" not in text) \
-                and ("%-" not in text) and ("%+" not in text):
-                midifile.text_events.recordText(track_desc.TotalClicksFromStart, text)
-            if debug:
-                print ("Text: %s" % (text))
+            if Length > 1000:
+                # This must be a mistake.
+                if debug:
+                    print ("Ignoring text of length %s" % (Length))
+            else:
+                if (midifile.text_encoding != "") :
+                    text = text.decode(midifile.text_encoding, 'replace')
+                # Take out any Sysex text events, and append to the lyrics list
+                if (" SYX" not in text) and ("Track-" not in text) \
+                    and ("%-" not in text) and ("%+" not in text):
+                    track_desc.text_events.recordText(track_desc.TotalClicksFromStart, text)
+                if debug:
+                    print ("Text: %s" % (repr(text)))
         elif event == 0x02:
             # Copyright (discard)
             Length, varBytes = varLength(filehdl)
@@ -639,7 +755,7 @@ def midiProcessEvent (filehdl, track_desc, midifile, ErrorNotifyCallback):
             title = filehdl.read(Length)
             bytesRead = bytesRead + Length
             if debug:
-                print ("Track Title: " + title)         
+                print ("Track Title: " + repr(title))
             if title == "Words":
                 track_desc.LyricsTrack = True
         elif event == 0x04:
@@ -653,18 +769,15 @@ def midiProcessEvent (filehdl, track_desc, midifile, ErrorNotifyCallback):
             Length, varBytes = varLength(filehdl)
             bytesRead = bytesRead + varBytes
             lyric = filehdl.read(Length)
-            if (midifile.text_charset != "") :
-                try:
-                    lyric = lyric.decode(midifile.text_charset)
-                except :
-                    print "Unicode conversion error"
+            if (midifile.text_encoding != "") :
+                lyric = lyric.decode(midifile.text_encoding, 'replace')
             bytesRead = bytesRead + Length
             # Take out any Sysex text events, and append to the lyrics list
             if (" SYX" not in lyric) and ("Track-" not in lyric) \
                 and ("%-" not in lyric) and ("%+" not in lyric):
-                midifile.lyric_events.recordLyric(track_desc.TotalClicksFromStart, lyric)
+                track_desc.lyric_events.recordLyric(track_desc.TotalClicksFromStart, lyric)
             if debug:
-                print ("Lyric: %s" % (lyric))
+                print ("Lyric: %s" % (repr(lyric)))
         elif event == 0x06:
             # Marker (discard)
             Length, varBytes = varLength(filehdl)
@@ -760,16 +873,18 @@ def midiProcessEvent (filehdl, track_desc, midifile, ErrorNotifyCallback):
                 print ("Manufacturer Data: " + data)
         else:
             # Unknown event (discard)
-            print ("Unknown meta-event: 0x%X" % event)
+            if debug:
+                print ("Unknown meta-event: 0x%X" % event)
             Length, varBytes = varLength(filehdl)
             bytesRead = bytesRead + varBytes
             discard = filehdl.read(Length)
             bytesRead = bytesRead + Length
 
     elif (event_type & 0xF0) == 0x80:
-        # Note off (discard)
+        # Note off
         packet = filehdl.read(2)
         bytesRead = bytesRead + 2
+        track_desc.LastNoteClick = track_desc.TotalClicksFromStart
     elif (event_type & 0xF0) == 0x90:
         # Note on (discard but note if the start time of the first in the track)
         packet = filehdl.read(2)
@@ -777,6 +892,7 @@ def midiProcessEvent (filehdl, track_desc, midifile, ErrorNotifyCallback):
         #print ("T%d: 0x%X" % (track_desc.TrackNum, event_type))
         if track_desc.FirstNoteClick == None:
             track_desc.FirstNoteClick = track_desc.TotalClicksFromStart
+        track_desc.LastNoteClick = track_desc.TotalClicksFromStart
     elif (event_type & 0xF0) == 0xA0:
         # Key after-touch (discard)
         packet = filehdl.read(2)
@@ -818,7 +934,8 @@ def midiProcessEvent (filehdl, track_desc, midifile, ErrorNotifyCallback):
         bytesRead = bytesRead + Length
     else:
         # Unknown event (discard)
-        print ("Unknown event: 0x%x" % event_type)
+        if debug:
+            print ("Unknown event: 0x%x" % event_type)
         Length, varBytes = varLength(filehdl)
         bytesRead = bytesRead + varBytes
         discard = filehdl.read(Length)
@@ -850,37 +967,40 @@ def varLength(filehdl):
 
 
 class midPlayer(pykPlayer):
-    def __init__(self, song, errorNotifyCallback=None, doneCallback=None, Charset="iso-8859-1"):
+    def __init__(self, song, songDb, errorNotifyCallback=None, doneCallback=None):
         """The first parameter, song, may be either a pykdb.SongStruct
         instance, or it may be a filename. """
 
-        pykPlayer.__init__(self, song, errorNotifyCallback, doneCallback)
+        pykPlayer.__init__(self, song, songDb, errorNotifyCallback, doneCallback)
+        settings = self.songDb.Settings
 
         self.SupportsFontZoom = True
+        self.isValid = False
 
         # Parse the MIDI file
-        self.midifile = midiParseData(self.SongDatas[0].GetData(), self.ErrorNotifyCallback, Charset)
+        self.midifile = midiParseData(self.SongDatas[0].GetData(), self.ErrorNotifyCallback, settings.KarEncoding)
         if (self.midifile == None):
             ErrorString = "ERROR: Could not parse the MIDI file"
             self.ErrorNotifyCallback (ErrorString)
             return
-        elif (self.midifile.lyrics == None):    
+        elif (self.midifile.lyrics == None):
             ErrorString = "ERROR: Could not get any lyric data from file"
             self.ErrorNotifyCallback (ErrorString)
             return
 
+        self.isValid = True
+
         # Debug out the found lyrics
         if debug:
             self.midifile.lyrics.write()
 
+        manager.setCpuSpeed('kar')
         manager.InitPlayer(self)
         manager.OpenDisplay()
 
-        # Reduce the default sample rate on the GP2x to save CPU time.
-        if env == ENV_GP2X:
-            manager.OpenAudio(suggestedProperties = (12000, -16, 1))
-        else:
-            manager.OpenAudio(suggestedProperties = (22050, -16, 2))
+        if not manager.options.nomusic:
+            manager.OpenAudio(frequency = manager.settings.MIDISampleRate,
+                              channels = 1)
 
         # Account for the size of the playback buffer in the lyrics
         # display.  Assume that the buffer will be mostly full.  On a
@@ -896,9 +1016,24 @@ class midPlayer(pykPlayer):
         # to the first note), so no need for the earliest note hack
         # there.  On timidity-based platforms, we anticipate our
         # lyrics display by the time of the first note.
+
+        # Note: pygame on OSX can run MIDI natively, or if the GUS
+        # patches are installed in /usr/local/lib/timidity, it will
+        # run MIDI via Timidity instead, which appears to work better
+        # than the native support, so we recommend this.
         if env != ENV_WINDOWS:
             self.InternalOffsetTime += self.midifile.earliestNoteMS
 
+        if env == ENV_LINUX:
+            # For some reason, a further offset of -250 ms seems about
+            # right empirically, on Linux, but not on Windows or on
+            # the GP2X.
+
+            # Actually, on further experimentation, this seems to be
+            # specific to my particular machine.  Leave it out.
+            #self.InternalOffsetTime -= 250
+            pass
+
         # Now word-wrap the text to fit our window.
         self.lyrics = self.midifile.lyrics.wordWrapLyrics(self.font)
 
@@ -908,23 +1043,28 @@ class midPlayer(pykPlayer):
         self.useMidiTimer = True
 
         if env == ENV_WINDOWS:
-            # Unless we're running on Windows.  For some reason, MIDI
-            # playback on Windows reports an unreliable time.  To
-            # avoid that problem, we'll always use the CPU timer
-            # instead of the MIDI timer.
+            # Unless we're running on Windows (i.e., not timidity).
+            # For some reason, hardware MIDI playback can report an
+            # unreliable time.  To avoid that problem, we'll always
+            # use the CPU timer instead of the MIDI timer.
             self.useMidiTimer = False
 
         # Load the MIDI player
-        if not manager.options.nomusic:
-            pygame.mixer.music.load(self.SongDatas[0].GetFilepath())
-
-            # Set an event for when the music finishes playing
-            pygame.mixer.music.set_endevent(pygame.USEREVENT)
-        else:
+        if manager.options.nomusic:
             # If we're not playing music, use the CPU timer instead of
             # the MIDI timer.
             self.useMidiTimer = False
 
+        else:
+            # Load the sound normally for playback.
+            audio_path = self.SongDatas[0].GetFilepath()
+            if type(audio_path) == unicode:
+                audio_path = audio_path.encode(sys.getfilesystemencoding())
+            pygame.mixer.music.load(audio_path)
+
+            # Set an event for when the music finishes playing
+            pygame.mixer.music.set_endevent(pygame.USEREVENT)
+
         # Reset all the state (current lyric index etc) and
         # paint the first numRows lines.
         self.resetPlayingState()
@@ -945,13 +1085,13 @@ class midPlayer(pykPlayer):
         # Remove irrelevant options.
         parser.remove_option('--fps')
         parser.remove_option('--zoom')
-        
+
         return parser
 
 
     def initFont(self):
         fontSize = int(FONT_SIZE * manager.GetFontScale() * manager.displaySize[1] / 480.)
-        self.font = pygame.font.Font(os.path.join(manager.FontPath, "DejaVuSans.ttf"), fontSize)
+        self.font = self.findPygameFont(self.songDb.Settings.KarFont, fontSize)
         self.lineSize = max(self.font.get_height(), self.font.get_linesize())
         self.numRows = int((manager.displaySize[1] - Y_BORDER * 2) / self.lineSize)
 
@@ -960,7 +1100,7 @@ class midPlayer(pykPlayer):
         self.viewRow = int(self.numRows * VIEW_PERCENT / 100)
 
     def resetPlayingState(self):
-    
+
         # Set the state variables
 
         # The current point the user was hearing within the song, as
@@ -995,9 +1135,10 @@ class midPlayer(pykPlayer):
 
     def repaintScreen(self):
         # Redraws the contents of the currently onscreen text.
-        
+
         # Clear the screen
-        manager.display.fill((0,0,0))
+        settings = self.songDb.Settings
+        manager.surface.fill(settings.KarBackgroundColour)
 
         # Paint the first numRows lines
         for i in range(self.numRows):
@@ -1009,7 +1150,7 @@ class midPlayer(pykPlayer):
                     self.drawSyllable(syllable, i, None)
                     x = syllable.right
 
-        pygame.display.flip()
+        manager.Flip()
         self.screenDirty = False
 
     def drawSyllable(self, syllable, row, x):
@@ -1021,37 +1162,54 @@ class midPlayer(pykPlayer):
         fill in the syllable's x position if it is not already known.
         x may be none if the syllable's x position is already
         known."""
-        
+
         if syllable.left == None:
             syllable.left = x
             if syllable.left == None:
                 return
-        
+
         y = Y_BORDER + row * self.lineSize
 
+        settings = self.songDb.Settings
+
         if syllable.type == TEXT_LYRIC:
-            # Lyrics start out red, and get painted white.
             if self.currentMs < syllable.ms:
-                color = (255,50,50)
+                color = settings.KarReadyColour
             else:
-                color = (255,255,255)
+                color = settings.KarSweepColour
         elif syllable.type == TEXT_INFO:
-            # Info messages are dark blue.
-            color = (0, 0, 200)
+            color = settings.KarInfoColour
         elif syllable.type == TEXT_TITLE:
-            # Titles are bright blue.
-            color = (100, 100, 255)
+            color = settings.KarTitleColour
 
         # Render text on a black background (instead of transparent)
         # to save a hair of CPU time.
-        text = self.font.render(syllable.text, True, color, (0, 0, 0))
-        
+        text = self.font.render(syllable.text, True, color,
+                                settings.KarBackgroundColour)
+
         width, height = text.get_size()
         syllable.right = syllable.left + width
 
-        manager.display.blit(text, (syllable.left, y, width, height))
+        manager.surface.blit(text, (syllable.left, y, width, height))
+
+    def __hasLyrics(self):
+        """ Returns true if the midi file contains any lyrics at all,
+        false if it doesn't (or contains only comments). """
+
+        if not self.midifile or not self.midifile.lyrics:
+            return False
+
+        for syllable in self.midifile.lyrics.list:
+            if syllable.type == TEXT_LYRIC:
+                return True
+        return False
+
+    def doValidate(self):
+        if not self.__hasLyrics():
+            return False
+
+        return True
 
- 
     def doPlay(self):
         if not manager.options.nomusic:
             pygame.mixer.music.play()
@@ -1062,14 +1220,14 @@ class midPlayer(pykPlayer):
             # front.
             pygame.time.wait(50)
 
-    def doPause(self): 
+    def doPause(self):
         if not manager.options.nomusic:
             pygame.mixer.music.pause()
 
     def doUnpause(self):
         if not manager.options.nomusic:
             pygame.mixer.music.unpause()
-        
+
     def doRewind(self):
         # Reset all the state (current lyric index etc)
         self.resetPlayingState()
@@ -1078,21 +1236,39 @@ class midPlayer(pykPlayer):
             pygame.mixer.music.rewind()
             pygame.mixer.music.stop()
 
+    def GetLength(self):
+        """Give the number of seconds in the song."""
+        return self.midifile.lastNoteMS / 1000
+
     def shutdown(self):
         # This will be called by the pykManager to shut down the thing
         # immediately.
         if not manager.options.nomusic:
-            pygame.mixer.music.stop()
+            if manager.audioProps:
+                pygame.mixer.music.stop()
         pykPlayer.shutdown(self)
 
 
     def doStuff(self):
         pykPlayer.doStuff(self)
 
-        if self.State == STATE_PLAYING:
-            self.currentMs = int(self.GetPos() + self.InternalOffsetTime + manager.UserOffsetTime)
+        if self.State == STATE_PLAYING or self.State == STATE_CAPTURING:
+            self.currentMs = int(self.GetPos() + self.InternalOffsetTime + manager.settings.SyncDelayMs)
             self.colourUpdateMs()
 
+            # If we're not using the automatic midi timer, we have to
+            # know to when stop the song at the end ourselves.
+            if self.currentMs > self.midifile.lastNoteMS:
+                self.Close()
+
+    def handleEvent(self, event):
+        if event.type == pygame.KEYDOWN and event.key == pygame.K_RETURN and (event.mod & (pygame.KMOD_LSHIFT | pygame.KMOD_RSHIFT | pygame.KMOD_LMETA | pygame.KMOD_RMETA)):
+            # Shift/meta return: start/stop song.  Useful for keybinding apps.
+            self.Close()
+            return
+        
+        pykPlayer.handleEvent(self, event)
+
     def doResize(self, newSize):
         # This will be called internally whenever the window is
         # resized for any reason, either due to an application resize
@@ -1135,7 +1311,7 @@ class midPlayer(pykPlayer):
                 self.drawSyllable(syllable, line - self.topLine, x)
                 x = syllable.right
 
-            pygame.display.flip()
+            manager.Flip()
 
         return True
 
@@ -1148,7 +1324,7 @@ class midPlayer(pykPlayer):
         self.nextSyllable, and self.nextColourMs. """
 
         syllables = []
-        
+
         while self.nextLine < len(self.lyrics):
             line = self.lyrics[self.nextLine]
             while self.nextSyllable < len(line):
@@ -1170,7 +1346,7 @@ class midPlayer(pykPlayer):
         # There are no more syllables to be displayed.
         self.nextColourMs = None
         return syllables
-        
+
 
     def considerScroll(self, syllables):
         """Determines whether it is time to scroll the screen.  If it
@@ -1178,7 +1354,7 @@ class midPlayer(pykPlayer):
         and returns the new list of syllables that need to be painted.
         If it is not yet time to scroll, does nothing and does not
         modify the syllable list. """
-        
+
         # If the player's still singing the top line, we can't scroll
         # it off yet.
         if self.currentLine <= self.topLine:
@@ -1224,14 +1400,15 @@ class midPlayer(pykPlayer):
         h = linesRemaining * self.lineSize
         rect = pygame.Rect(X_BORDER, y,
                            manager.displaySize[0] - X_BORDER * 2, h)
-        manager.display.blit(manager.display, (X_BORDER, Y_BORDER), rect)
+        manager.surface.blit(manager.surface, (X_BORDER, Y_BORDER), rect)
 
         # And now fill the lower part of the screen with black.
         y = Y_BORDER + linesRemaining * self.lineSize
         h = linesScrolled * self.lineSize
         rect = pygame.Rect(X_BORDER, y,
                            manager.displaySize[0] - X_BORDER * 2, h)
-        manager.display.fill((0, 0, 0), rect)
+        settings = self.songDb.Settings
+        manager.surface.fill(settings.KarBackgroundColour, rect)
 
         # We can remove any syllables from the list that might have
         # scrolled off the screen now.
@@ -1250,7 +1427,7 @@ class midPlayer(pykPlayer):
                 syllables.append((syllable, i))
 
         return syllables
-        
+
 
 def usage():
     print "Usage:  %s <kar filename>" % os.path.basename(sys.argv[0])
@@ -1258,9 +1435,10 @@ def usage():
 
 # Can be called from the command line with the CDG filepath as parameter
 def main():
-    player = midPlayer(None)
-    player.Play()
-    manager.WaitForPlayer()
+    player = midPlayer(None, None)
+    if player.isValid:
+        player.Play()
+        manager.WaitForPlayer()
 
 if __name__ == "__main__":
     sys.exit(main())
