@@ -1,7 +1,7 @@
 #******************************************************************************
 #****                                                                      ****
-#**** Copyright (C) 2007  Kelvin Lawson (kelvinl@users.sourceforge.net)    ****
-#**** Copyright (C) 2009  PyKaraoke Development Team                       ****
+#**** Copyright (C) 2010  Kelvin Lawson (kelvinl@users.sourceforge.net)    ****
+#**** Copyright (C) 2010  PyKaraoke Development Team                       ****
 #****                                                                      ****
 #**** This library is free software; you can redistribute it and/or        ****
 #**** modify it under the terms of the GNU Lesser General Public           ****
@@ -28,9 +28,12 @@ from pykconstants import *
 from pykenv import env
 import pykar, pycdg, pympg
 import os, cPickle, zipfile, codecs, sys, time
-import md5
 import types
 from cStringIO import StringIO
+try:
+    from hashlib import md5
+except ImportError:
+    from md5 import md5
 
 # The amount of time to wait, in milliseconds, before yielding to the
 # app for windowing updates during a long update process.
@@ -43,7 +46,7 @@ MAX_ZIP_FILES = 10
 # Increment this version number whenever the settings version changes
 # (which may not necessarily change with each PyKaraoke release).
 # This will force users to re-enter their configuration information.
-SETTINGS_VERSION = 5
+SETTINGS_VERSION = 6
 
 # Increment this version number whenever the database version changes
 # (which will also hopefully be infrequently).
@@ -137,7 +140,12 @@ class SongData:
             # to a temporary file.
             prefix = globalSongDB.CreateTempFileNamePrefix()
             basename = os.path.basename(self.filename)
-            self.tempFilename = prefix + basename
+            # Add the tempfile prefix as well as the time to make the
+            # filename unique. This works around a bug
+            # in pygame.mixer.music on Windows which stops us deleting
+            # temp files until another song is loaded, preventing the
+            # same song from being played twice in a row.
+            self.tempFilename = prefix + str(time.time()) + basename
             open(self.tempFilename, 'wb').write(self.data)
 
         return self.tempFilename
@@ -160,20 +168,36 @@ class SongStruct:
     T_MPG = 2
 
     def __init__(self, Filepath, settings,
-                 Title = None, Artist = None, ZipStoredName = None):
+                 Title = None, Artist = None, ZipStoredName = None, DatabaseAdd = False):
         self.Filepath = Filepath    # Full path to file or ZIP file
         self.ZipStoredName = ZipStoredName # Filename stored in ZIP
-        # Check to see if we are deriving song information
+
+        # Assume there will be no title/artist info found
+        self.Title = Title or ''    # (optional) Title for display in playlist
+        self.Artist = Artist or ''  # (optional) Artist for display
+        self.Disc = '' # (optional) Disc for display
+        self.Track = -1 # (optional) Track for display
+
+        # Check to see if we are deriving song information from the filename
         if settings.CdgDeriveSongInformation:
-            self.Title = self.ParseTitle(Filepath, settings)    # Title for display in playlist
-            self.Artist = self.ParseArtist(Filepath, settings)  # Artist for display
-            self.Disc = self.ParseDisc(Filepath, settings) # Disc for display
-            self.Track = self.ParseTrack(Filepath, settings) # Track for display
-        else:
-            self.Title = Title or ''    # (optional) Title for display in playlist
-            self.Artist = Artist or ''  # (optional) Artist for display
-            self.Disc = '' # (optional) Disc for display
-            self.Track = -1 # (optional) Track for display
+            try:
+                self.Title = self.ParseTitle(Filepath, settings)    # Title for display in playlist
+                self.Artist = self.ParseArtist(Filepath, settings)  # Artist for display
+                self.Disc = self.ParseDisc(Filepath, settings)      # Disc for display
+                self.Track = self.ParseTrack(Filepath, settings)     # Track for display
+            except:
+                # Filename did not match requested scheme, set the title to the filepath
+                # so that the structure is still created, but without any additional info
+                #print "Filename format does not match requested scheme: %s" % Filepath
+                self.Title = os.path.basename(Filepath)
+
+                # If this SongStruct is being used to add to the database, and we are
+                # configured to exclude non-matching files, raise an exception. Otherwise
+                # allow it through. For database adds where we are not excluding such
+                # files the song will still be added to the database. For non-database
+                # adds we don't care anyway, we just want a SongStruct for passing around.
+                if DatabaseAdd and settings.ExcludeNonMatchingFilenames:
+                    raise KeyError, "Excluding non-matching file: %s" % self.Title
 
         # This is a list of other song files that share the same
         # artist and title data.
@@ -312,6 +336,25 @@ class SongStruct:
         #print "Track parsed: %s" % track
         return track
 
+    def MakeSortKey(self, str):
+        """ Returns a suitable key to use for sorting, by lowercasing
+        and removing articles from the indicated string. """
+        str = str.strip().lower()
+        if str:
+            # Remove a leading parenthetical phrase.
+            if str[0] == '(':
+                rparen = str.index(')')
+                if rparen != ')':
+                    str = str[rparen + 1:].strip()
+
+        if str:
+            # Remove a leading article.
+            firstWord = str.split()[0]
+            if firstWord in ['a', 'an', 'the']:
+                str = str[len(firstWord):].strip()
+                
+        return str
+
     def MakePlayer(self, songDb, errorNotifyCallback, doneCallback):
         """Creates and returns a player of the appropriate type to
         play this file, if possible; or returns None if the file
@@ -482,6 +525,11 @@ class SongStruct:
         # files in the mini index.
         return (-self.Type, self.DisplayFilename)
 
+    def getMarkKey(self):
+        """ Returns a key for indexing into markedSongs, for uniquely
+        identifying this particular song file. """
+        return (self.Filepath, self.ZipStoredName)
+
     def __cmp__(self, other):
         """Define a sorting order between SongStruct objects.  This is
         used in bisect, to quickly search for a SongStruct in a sorted
@@ -592,7 +640,12 @@ class TitleStruct:
                 return
 
         for line in catalogFile:
-            line = line.decode('utf-8').strip()
+            try:
+                line = line.decode('utf-8').strip()
+            except UnicodeDecodeError:
+                line = line.decode('utf-8', 'replace')
+                print "Invalid characters in %s:\n%s" % (catalogPathname, line)
+
             if line:
                 tuple = line.split('\t')
                 if len(tuple) == 2:
@@ -785,8 +838,15 @@ class SettingsStruct:
         self.WindowSize = (640, 480) # Size of the window for PyKaraoke
         self.FullScreen = False # Determines if the karaoke player should be full screen
         self.NoFrame = False # Determies if the karaoke player should have a window frame.
+
+        # SDL specific parameters; some settings may work better on
+        # certain hardware than others
+        self.DoubleBuf = True
+        self.HardwareSurface = True
+        
         self.PlayerSize = (640, 480) # Size of the karaoke player
         self.PlayerPosition = None # Initial position of the karaoke player
+        
         self.SplitVertically = True
         self.AutoPlayList = True # Enables or disables the auto play on the play-list
         self.DoubleClickPlayList = True # Enables or disables the double click for playing from the play-list
@@ -809,6 +869,7 @@ class SettingsStruct:
         # negative values delay them.
         self.SyncDelayMs = 0
 
+        # KAR/MID options
         self.KarEncoding = 'cp1252'  # Default text encoding in karaoke files
         self.KarFont = FontData("DejaVuSans.ttf")
         self.KarBackgroundColour = (0, 0, 0)
@@ -818,11 +879,14 @@ class SettingsStruct:
         self.KarTitleColour = (100, 100, 255)
         self.MIDISampleRate = 44100
 
+        # CDG options
         self.CdgZoom = 'int'
         self.CdgUseC = True
         self.CdgDeriveSongInformation = False # Determines if we should parse file names for song information
         self.CdgFileNameType = -1 # The style index we are using for the file name parsing
+        self.ExcludeNonMatchingFilenames = False # Exclude songs from database if can't derive song info
 
+        # MPEG options
         self.MpgNative = True
         self.MpgExternalThreaded = True
         self.MpgExternal = 'mplayer -fs "%(file)s"'
@@ -1271,7 +1335,7 @@ class SongDB:
             # going to traverse.  We give each directory equal weight
             # regardless of the number of files within it.
             progress = [(i, len(fileList))]
-            self.fileScan(str(root_path), progress, yielder)
+            self.fileScan(root_path, progress, yielder)
             if self.BusyDlg.Clicked:
                 break
 
@@ -1366,7 +1430,7 @@ class SongDB:
             # Every so often, update the progress bar.
             basename = os.path.split(full_path)[1]
             self.BusyDlg.SetProgress(
-                "Scanning %s" % unicode(basename, errors='replace'),
+                "Scanning %s" % basename,
                 self.__computeProgressValue(progress))
             yielder.Yield()
             self.lastBusyUpdate = now
@@ -1384,46 +1448,54 @@ class SongDB:
         # Store file details if it's a file type we're interested in
         else:
             root, ext = os.path.splitext(full_path)
-            try:
-                # Non-ZIP files
-                if self.Settings.ReadTitlesTxt and full_path.endswith('titles.txt'):
-                    # Save this titles.txt file for reading later.
-                    self.TitlesFiles.append(TitleStruct(full_path))
-                elif self.IsExtensionValid(ext):
-                    self.addSong(SongStruct(full_path, self.Settings))
-                # Look inside ZIPs if configured to do so
-                elif self.Settings.LookInsideZips and ext.lower() == ".zip":
-                        if zipfile.is_zipfile(full_path):
-                            zip = self.GetZipFile(full_path)
-                            namelist = zip.namelist()
-                            for i in range(len(namelist)):
-                                filename = namelist[i]
+            # Non-ZIP files
+            if self.Settings.ReadTitlesTxt and full_path.endswith('titles.txt'):
+                # Save this titles.txt file for reading later.
+                self.TitlesFiles.append(TitleStruct(full_path))
+            elif self.IsExtensionValid(ext):
+                try:
+                    self.addSong(SongStruct(full_path, self.Settings, DatabaseAdd = True))
+                except KeyError:
+                    print "Excluding filename with unexpected format: %s " % os.path.basename(full_path)
+            # Look inside ZIPs if configured to do so
+            elif self.Settings.LookInsideZips and ext.lower() == ".zip":
+                try:
+                    if zipfile.is_zipfile(full_path):
+                        zip = self.GetZipFile(full_path)
+                        namelist = zip.namelist()
+                        for i in range(len(namelist)):
+                            filename = namelist[i]
 
-                                now = time.time()
-                                if now - self.lastBusyUpdate > 0.1:
-                                    # Every so often, update the progress bar.
-                                    nextProgress = progress + [(i, len(namelist))]
-                                    basename = os.path.split(full_path)[1]
-                                    self.BusyDlg.SetProgress(
-                                        "Scanning %s" % (basename),
-                                        self.__computeProgressValue(nextProgress))
-                                    yielder.Yield()
-                                    self.lastBusyUpdate = now
+                            now = time.time()
+                            if now - self.lastBusyUpdate > 0.1:
+                                # Every so often, update the progress bar.
+                                nextProgress = progress + [(i, len(namelist))]
+                                basename = os.path.split(full_path)[1]
+                                self.BusyDlg.SetProgress(
+                                    "Scanning %s" % (basename),
+                                    self.__computeProgressValue(nextProgress))
+                                yielder.Yield()
+                                self.lastBusyUpdate = now
 
-                                root, ext = os.path.splitext(filename)
-                                if self.Settings.ReadTitlesTxt and filename.endswith('titles.txt'):
-                                    # Save this titles.txt file for reading later.
-                                    self.TitlesFiles.append(TitleStruct(full_path, ZipStoredName = filename))
-                                elif self.IsExtensionValid(ext):
-                                    # Python zipfile only supports deflated and stored
-                                    info = zip.getinfo(filename)
-                                    if info.compress_type == zipfile.ZIP_STORED or info.compress_type == zipfile.ZIP_DEFLATED:
-                                        #print ("Adding song %s in ZIP file %s"%(filename, full_path))
-                                        self.addSong(SongStruct(full_path, self.Settings, ZipStoredName = filename))
-                                    else:
-                                        print ("ZIP member %s compressed with unsupported type (%d)"%(full_path, info.compress_type))
-            except:
-                print "Error looking inside zip " + full_path
+                            root, ext = os.path.splitext(filename)
+                            if self.Settings.ReadTitlesTxt and filename.endswith('titles.txt'):
+                                # Save this titles.txt file for reading later.
+                                self.TitlesFiles.append(TitleStruct(full_path, ZipStoredName = filename))
+                            elif self.IsExtensionValid(ext):
+                                # Python zipfile only supports deflated and stored
+                                info = zip.getinfo(filename)
+                                if info.compress_type == zipfile.ZIP_STORED or info.compress_type == zipfile.ZIP_DEFLATED:
+                                    #print ("Adding song %s in ZIP file %s"%(filename, full_path))
+                                    try:
+                                        self.addSong(SongStruct(full_path, self.Settings, ZipStoredName = filename, DatabaseAdd = True))
+                                    except KeyError:
+                                        print "Excluding filename with unexpected format: %s " % os.path.basename(full_path)
+                                else:
+                                    print ("ZIP member compressed with unsupported type (%d): %s"%(info.compress_type, full_path))
+                    else:
+                        print "Cannot parse ZIP file: " + full_path 
+                except:
+                    print "Error looking inside zip " + full_path
 
     # Add a folder to the database search list
     def FolderAdd (self, FolderPath):
@@ -1512,71 +1584,92 @@ class SongDB:
                     try:
                         os.unlink(full_path)
                     except:
+                        # The unlink can fail on Windows due to a bug in
+                        # pygame.mixer.music which does not release the
+                        # file handle until you load another music file.
                         pass
 
-    def SelectSort(self, sort):
+    def SelectSort(self, sort, allowResort = True):
         """Sorts the list of songs in order according to the indicated
         key, which must be one of 'title', 'artist', or 'filename'.
         Also sets self.GetSongTuple to a functor which, when called, returns a
         3-tuple of strings suitable for displaying for each song,
-        where the first string of each tuple is the sort key. """
+        where the first string of each tuple is the sort key.
 
-        self.Sort = sort
-        global fileSortKey
+        This may require re-sorting the list on-the-fly.  If
+        allowResort is False, then the list will never be re-sorted;
+        rather, the method will return True if the sort was
+        successfully applied, or False if a re-sort was necessary but
+        not performed, in which case the list retains its original
+        sort. """
 
-        if self.Sort == 'title' and self.GotTitles:
+        if sort == 'title' and self.GotTitles:
             if self.GotArtists:
-                self.GetSongTuple = self.getSongTupleTitleArtistFilename
-                self.SortKeys = ('title', 'artist', 'filename')
-                fileSortKey = self.getSongTupleTitleArtistFilenameSortKey
+                getSongTuple = self.getSongTupleTitleArtistFilename
+                sortKeys = ('title', 'artist', 'filename')
+                getSortKey = self.getSongTupleTitleArtistFilenameSortKey
             else:
-                self.GetSongTuple = self.getSongTupleTitleFilenameArtist
-                self.SortKeys = ('title', 'filename')
-                fileSortKey = self.getSongTupleTitleFilenameArtistSortKey
+                getSongTuple = self.getSongTupleTitleFilenameArtist
+                sortKeys = ('title', 'filename')
+                getSortKey = self.getSongTupleTitleFilenameArtistSortKey
 
-        elif self.Sort == 'artist' and self.GotArtists:
+        elif sort == 'artist' and self.GotArtists:
             if self.GotTitles:
-                self.GetSongTuple = self.getSongTupleArtistTitleFilename
-                self.SortKeys = ('artist', 'title', 'filename')
-                fileSortKey = self.getSongTupleArtistTitleFilenameSortKey
+                getSongTuple = self.getSongTupleArtistTitleFilename
+                sortKeys = ('artist', 'title', 'filename')
+                getSortKey = self.getSongTupleArtistTitleFilenameSortKey
             else:
-                self.GetSongTuple = self.getSongTupleArtistFilenameTitle
-                self.SortKeys = ('artist', 'filename')
-                fileSortKey = self.getSongTupleArtistFilenameTitleSortKey
+                getSongTuple = self.getSongTupleArtistFilenameTitle
+                sortKeys = ('artist', 'filename')
+                getSortKey = self.getSongTupleArtistFilenameTitleSortKey
 
         else: # filename
-            self.Sort = 'filename'
+            sort = 'filename'
             if self.GotTitles and self.GotArtists:
-                self.GetSongTuple = self.getSongTupleFilenameTitleArtist
-                self.SortKeys = ('filename', 'title', 'artist')
-                fileSortKey = self.getSongTupleFilenameTitleArtistSortKey
+                getSongTuple = self.getSongTupleFilenameTitleArtist
+                sortKeys = ('filename', 'title', 'artist')
+                getSortKey = self.getSongTupleFilenameTitleArtistSortKey
             elif self.GotTitles:
-                self.GetSongTuple = self.getSongTupleFilenameTitleArtist
-                self.SortKeys = ('filename', 'title')
-                fileSortKey = self.getSongTupleFilenameTitleArtistSortKey
+                getSongTuple = self.getSongTupleFilenameTitleArtist
+                sortKeys = ('filename', 'title')
+                getSortKey = self.getSongTupleFilenameTitleArtistSortKey
             elif self.GotArtists:
-                self.GetSongTuple = self.getSongTupleFilenameArtistTitle
-                self.SortKeys = ('filename', 'artist')
-                fileSortKey = self.getSongTupleFilenameArtistTitleSortKey
+                getSongTuple = self.getSongTupleFilenameArtistTitle
+                sortKeys = ('filename', 'artist')
+                getSortKey = self.getSongTupleFilenameArtistTitleSortKey
             else:
-                self.GetSongTuple = self.getSongTupleFilenameArtistTitle
-                self.SortKeys = ('filename',)
-                fileSortKey = self.getSongTupleFilenameArtistTitleSortKey
+                getSongTuple = self.getSongTupleFilenameArtistTitle
+                sortKeys = ('filename',)
+                getSortKey = self.getSongTupleFilenameArtistTitleSortKey
 
-        list = self.SortedLists.get(fileSortKey, None)
+        list = self.SortedLists.get(getSortKey, None)
         if list is None:
+            if not allowResort:
+                # Return False to indicate that a sort was not applied.
+                return False
+            
             # We haven't asked for this sort key before; we have to
             # sort the list now.  Once sorted, we can keep it around
             # for future requests.
-            if self.Sort == 'filename':
+            if sort == 'filename':
                 list = self.FullSongList[:]
             else:
                 list = self.UniqueSongList[:]
-            list.sort(key = fileSortKey)
-            self.SortedLists[fileSortKey] = list
+            list.sort(key = getSortKey)
+            self.SortedLists[getSortKey] = list
+
+        global fileSortKey
+        fileSortKey = getSortKey
+
+        self.Sort = sort
+        self.SortKeys = sortKeys
+        self.GetSongTuple = getSongTuple
+        self.GetSortKey = getSortKey
 
         self.SongList = list
 
+        return True
+
     def getSongTupleTitleArtistFilename(self, file):
         return (file.Title, file.Artist, file.getDisplayFilenames())
 
@@ -1596,22 +1689,22 @@ class SongDB:
         return (file.DisplayFilename, file.Artist, file.Title)
 
     def getSongTupleTitleArtistFilenameSortKey(self, file):
-        return (file.Title.lower(), file.Artist.lower(), file.getDisplayFilenames().lower(), id(file))
+        return (file.MakeSortKey(file.Title), file.MakeSortKey(file.Artist), file.getDisplayFilenames().lower(), id(file))
 
     def getSongTupleTitleFilenameArtistSortKey(self, file):
-        return (file.Title.lower(), file.DisplayFilename.lower(), file.Artist.lower(), id(file))
+        return (file.MakeSortKey(file.Title), file.DisplayFilename.lower(), file.MakeSortKey(file.Artist), id(file))
 
     def getSongTupleArtistTitleFilenameSortKey(self, file):
-        return (file.Artist.lower(), file.Title.lower(), file.DisplayFilename.lower(), id(file))
+        return (file.MakeSortKey(file.Artist), file.MakeSortKey(file.Title), file.DisplayFilename.lower(), id(file))
 
     def getSongTupleArtistFilenameTitleSortKey(self, file):
-        return (file.Artist.lower(), file.DisplayFilename.lower(), file.Title.lower(), id(file))
+        return (file.MakeSortKey(file.Artist), file.DisplayFilename.lower(), file.MakeSortKey(file.Title), id(file))
 
     def getSongTupleFilenameTitleArtistSortKey(self, file):
-        return (file.DisplayFilename.lower(), file.Title.lower(), file.Artist.lower(), id(file))
+        return (file.DisplayFilename.lower(), file.MakeSortKey(file.Title), file.MakeSortKey(file.Artist), id(file))
 
     def getSongTupleFilenameArtistTitleSortKey(self, file):
-        return (file.DisplayFilename.lower(), file.Artist.lower(), file.Title.lower(), id(file))
+        return (file.DisplayFilename.lower(), file.MakeSortKey(file.Artist), file.MakeSortKey(file.Title), id(file))
 
     def addSong(self, file):
         self.FullSongList.append(file)
@@ -1659,7 +1752,7 @@ class SongDB:
                 return
 
             song = self.FullSongList[i]
-            m = md5.md5()
+            m = md5()
             m.update(song.GetSongDatas()[0].GetData())
             list = fileHashes.setdefault(m.digest(), [])
             if list:
